//S.java - S(u,v) selection block
//Copyright (c) 2007 Thomas Schneider
//See full Copyright license terms in file ../GPL.txt

package SFE.PF;

import org.apache.log4j.Logger;

/**
 * S(u,v) selection block.
 * <p>
 * Select for each of the v outputs one of the u inputs. (with duplicates)
 *
 * @author Thomas Schneider
 */
abstract public class S extends Block {
	protected static final Logger logger = Logger.getLogger(S.class);

	/**
	 * Program selection block.
	 * 
	 * @param s selection mapping: output[i] <- input[s[i]], i=0..(v-1)
	 */
	abstract public void program(int[] s);

	/**
	 * Generate selection block.
	 * 
	 * @param u #inputs
	 * @param v #outputs
	 * @param c UC containing this selection block
	 */
	abstract public void generate(int u, int v, UC c);
	
	/**
	 * Estimate number of gates of selection block.
	 * 
	 * @param u #inputs
	 * @param v #outputs
	 * @return estimated #gates of selection block
	 */
	abstract public int estimateGates(int u, int v);
	
	/**
	 * Generate smallest selection block for given parameters.
	 * 
	 * @param u #inputs
	 * @param v #outputs
	 * @param c UC containing the selection block
	 * @return generated selection block
	 */
	public static S generateS(int u, int v, UC c) {
		String Snames[]={"S_u_v","S_u_2u","S_U","S_V"};
		int Ssize[]=new int[Snames.length];
		int minIndex=0;
		S aktS=null, minS=null;

		for(int i=0; i<Snames.length; i++) {
			// logger.info("Estimating size for "+Snames[i]);

			// load class
			String S_type="SFE.PF."+Snames[i];
			try{
				aktS=(S)java.lang.ClassLoader.getSystemClassLoader().loadClass(S_type).newInstance();
			}
			catch(Exception e) {
				e.printStackTrace();
				logger.error("Unable to load "+S_type);
				System.exit(1);
			}

			// determine size
			Ssize[i]=aktS.estimateGates(u,v);
			if(i==0 || Ssize[i]<Ssize[minIndex]) {
				minIndex=i;
				minS=aktS;
			}
		}
		String log="S("+u+","+v+")="+Snames[minIndex]+" [";
		for(int i=0; i<Snames.length; i++) {
			if(Ssize[i]==Integer.MAX_VALUE) log+=" -";
			else log+=" "+Ssize[i];
		}
		log+=" ] [";
		for(int i=0; i<Snames.length; i++)
			log+=" "+Snames[i];
		log+=" ]";
		logger.info(log);
		
		minS.generate(u, v, c);
		return minS;
	}
	
	/**
	 * Constructor
	 * 
	 * @param u #inputs
	 * @param v #outputs
	 * @param c UC containing this selection block
	 */
	public S(int u, int v, UC c) {
		super(u,v,c);
		generate(u,v,c);
	}
	
	/**
	 * Dummy constructor
	 */
	S() {
		super();
	}
}
