//UC.java - abstraction for universal circuits
//Copyright (c) 2007 Thomas Schneider
//See full Copyright license terms in file ../GPL.txt

package SFE.PF;

import org.apache.log4j.Logger;
import SFE.BOAL.Circuit;
import SFE.BOAL.Gate;

/**
 * A Universal Circuit (UC) is a circuit that can be programmed to simulate any circuit.
 *
 * @author Thomas Schneider
 */
public abstract class UC extends Circuit {
	protected static final Logger logger = Logger.getLogger(UC.class);

	protected int u;	/* #inputs */
	protected int v;	/* #outputs */
	protected int k;	/* #gates to simulate */

	protected boolean is_generated=false;	/* is this UC generated? */
	protected boolean is_implemented=false;	/* is this UC fully implemented? */
	
	protected int output_indexes[];	/* indexes of output gates */
	protected Gate outputs[];		/* output gates */
	
	/**
	 * Program UC with given circuit.
	 * 
	 * @param ins inputs
	 * @param outs outputs
	 * @param gates gates of circuit
	 */
	public abstract void program(int ins[], int outs[], ParseGate gates[]);
	
	/**
	 * Generate UC for circuit of given shape.
	 * 
	 * @param u #inputs
	 * @param v #outputs
	 * @param k #gates
	 */
	public abstract void generate(int u, int v, int k);
	
	/**
	 * Extract circuit representation of UC.
	 * 
	 * @return output gates
	 */
	public abstract Gate[] extractCircuit(Gate[] outputs);
	
	/**
	 * Count number of gates of UC.
	 * 
	 * @return #gates of UC
	 */
	public abstract int countGates();
	
	/**
	 * Estimate number of gates of UC.
	 * 
	 * @param u #inputs
	 * @param v #outputs
	 * @param k #gates of circuit
	 * @return estimated #gates of UC
	 */
	public abstract int estimateGates(int u, int v, int k);

	// -----------------------------------------
	public void generateEncCircuit() {
		logger.info("Encrypting circuit.");
		super.generateEncCircuit();
	}
	
	private int nextAliceInput, nextBobInput;
	private int nextAliceOutput, nextBobOutput;
	private boolean first_translateIO=true;
	protected int translateIn[], translateOut[];

	/**
	 * Translate inputs and outputs.
	 * <p>
	 * Inputs and outputs (IO) are rearranged to contain Alice's first and then Bob's in reversed order.
	 * Otherwise Alice could infer information about Bob's IO from the UC.
	 * <p>
	 * @param linenum original index of IO
	 * @param isAlice IO of Alice?
	 * @param isInput Input?
	 * @return translated IO index
	 */
	public int translateIO(int linenum, boolean isAlice, boolean isInput) {
		// init
		if(first_translateIO) {
			first_translateIO=false;
			nextAliceInput=0;
			nextBobInput=u-1;
			nextAliceOutput=0;
			nextBobOutput=v-1;
			translateIn=new int[u];
			translateOut=new int[v];
		}		
		if(isAlice) {
			if(isInput) {
		        // Alice's input
				translateIn[linenum]=nextAliceInput;
				return nextAliceInput++;
			}
			else {
		        // Alice's output
				translateOut[nextAliceOutput]=linenum;
				return output_indexes[nextAliceOutput++];
			}
		}
		else {
			if(isInput) {
		        // Bob's input
				translateIn[linenum]=nextBobInput;
				return nextBobInput--;
			}
			else {
		        // Bob's output
				translateOut[nextBobOutput]=linenum;
				return output_indexes[nextBobOutput--];
			}
		}
	}
	
	// -----------------------------------------
	/**
	 * generate UC for given circuit parameters
	 * 
	 * @param u #inputs
	 * @param v #outputs
	 * @param k #gates
	 * @param type type of UC, choose smallest UC if type=null
	 */
	public static UC generateUC(int u, int v, int k, String type) {
		UC c=null;
		int cGates=Integer.MAX_VALUE;
		String UC_name=null;
		if(type!=null) { // use given UC
			UC_name=type;
			type="SFE.PF."+type;
			try{
				c=(UC)java.lang.ClassLoader.getSystemClassLoader().loadClass(type).newInstance();
			}
			catch(Exception e) {
				e.printStackTrace();
				System.exit(1);
			}
		}
		else { // choose best UC
			String Unames[]={"UCsel","UCpract","UCvaliant"};
			int Usize[]=new int[Unames.length];
			UC Uakt=null, Umin=null;
			int minIndex=0;
			
			for(int i=0; i<Unames.length; i++) {
				// load class
				try{
					type="SFE.PF."+Unames[i];
					Uakt=(UC)java.lang.ClassLoader.getSystemClassLoader().loadClass(type).newInstance();
				}
				catch(Exception e) {
					logger.debug("Unable to load "+type+". Skipping.");
					Usize[i]=Integer.MAX_VALUE;
					continue;
				}

				// determine size
				Usize[i]=Uakt.estimateGates(u,v,k);
				if(Usize[i]==Integer.MAX_VALUE) {
					Uakt.generate(u,v,k);
					Usize[i]=Uakt.countGates();					
				}
				
				if(i==0 || Usize[i] < Usize[minIndex]) {
					if(Uakt.is_implemented) {
						minIndex=i;
						Umin=Uakt;
					}
				}
			}
			
			// log						
			logger.info("Searching optimal UC for u="+u+", v="+v+", k="+k+":");
			String log;
			// find maxlength
			int namelen=0;
			int sizelen=1;
			for(int i=0; i<Unames.length; i++) {
				  if(Unames[i].length()>namelen) namelen=Unames[i].length();
				  if(Usize[i]!=Integer.MAX_VALUE) {
					  int len=Integer.toString(Usize[i]).length();
					  if(len>sizelen) sizelen=len;
				  }
			}
			for(int i=0; i<Unames.length; i++) {
				log="  "+Unames[i];
				for(int j=Unames[i].length(); j<namelen; j++) log+=" ";
				log+=": ";
				if(Usize[i]==Integer.MAX_VALUE) {
					for(int j=1;j<sizelen;j++) log+=" ";
					log+="-";
				}
				else {
					for(int j=Integer.toString(Usize[i]).length();j<sizelen;j++) log+=" ";
					log+=Usize[i];
				}
				logger.info(log);
			}
			
			c=Umin;
			cGates=Usize[minIndex];
			UC_name=Unames[minIndex];
		}
		if(!c.is_generated) {
			c.generate(u,v,k);
		}
		if(cGates==Integer.MAX_VALUE) cGates=c.countGates();

		logger.info("Using "+UC_name+" with "+cGates+" gates.");
		
		return c;
	}
	
	// -----------------------------------------
	/** number of gates to implement a Y switching block */
	public static final int Ygates=1;
	
	/** number of gates to implement a X switching block */
	public static final int Xgates=2;

	/** number of gates to implement a gate simulation block */
	public static final int Ggates=1;
	
	/**
	 * 1 output switching gate
	 */
	protected class Y {
		Gate g;

		void program(boolean right) {
			//logger.debug(g+": "+(right?"R":"L"));
			if(!right) {
				g.setOutput(0, false);
				g.setOutput(1, true);
				g.setOutput(2, false);
				g.setOutput(3, true);
			}
			else {
				g.setOutput(0, false);
				g.setOutput(1, false);
				g.setOutput(2, true);
				g.setOutput(3, true);
			}
		}
				
		Gate extractCircuit(Gate input1, Gate input2) {
			g=addGate(2, Gate.REG_GATE, false);
			g.setInput(0, input1);
			g.setInput(1, input2);
			//logger.debug("Y: "+g);
			return g;
		}
	}
	
	/**
	 * 2 output switching gate
	 */
	protected class X {
		Gate g0, g1;

		void program(boolean sw) {
			if (!sw) {
				g0.setOutput(0, false);
				g0.setOutput(1, true);
				g0.setOutput(2, false);
				g0.setOutput(3, true);

				g1.setOutput(0, false);
				g1.setOutput(1, false);
				g1.setOutput(2, true);
				g1.setOutput(3, true);
			}
			else {
				g0.setOutput(0, false);
				g0.setOutput(1, false);
				g0.setOutput(2, true);
				g0.setOutput(3, true);

				g1.setOutput(0, false);
				g1.setOutput(1, true);
				g1.setOutput(2, false);
				g1.setOutput(3, true);
			}
		}
		
		Gate[] extractCircuit(Gate input1, Gate input2, Gate[] outtmp) {
			Gate[] g;
			if(outtmp!=null) g=outtmp;
			else g=new Gate[2];
			g0=addGate(2, Gate.REG_GATE, false);
			g0.setInput(0, input1);
			g0.setInput(1, input2);
			g1=addGate(2, Gate.REG_GATE, false);
			g1.setInput(0, input1);
			g1.setInput(1, input2);
			//logger.debug("X: "+g0+", "+g1);
			g[0]=g0; g[1]=g1;
			return g;
		}
	}

	/**
	 * gate that simulates a gate of the original circuit
	 */
	protected class G {
		Gate g;
		
		void program(int truthTable[]) {
			//logger.debug(g+": ["+truthTable[0]+" "+truthTable[1]+" "+truthTable[2]+" "+truthTable[3]+"]");
			for(int i=0; i<4; i++)
				g.setOutput(i, truthTable[i]==1);
		}
		
		Gate extractCircuit(Gate input1, Gate input2) {
			g=addGate(2, Gate.REG_GATE, false);
			g.setInput(0, input1);
			g.setInput(1, input2);
			//logger.debug("G: "+g);
			return g;
		}
	}
}
