//UCsel.java - Universal Circuit based on selection blocks
//Copyright (c) 2007 Thomas Schneider
//See full Copyright license terms in file ../GPL.txt

package SFE.PF;

import SFE.BOAL.Gate;

/**
 * Universal Circuit based on selection blocks.
 * <p>
 * sizeof(UCsel) = k^2 + k ( 2 u + v - 2 ) - v
 * 
 * As this UC construction grows quadratically in k it is only useful for k < 1000 !
 * However it is good for demonstration purposes because of very simple construction.
 *
 * For the i-th gate (i=1..k) select for each of the two gate inputs the corresponding input value
 * with a selection block that selects from all inputs and the outputs of previous k g < i.
 * For each output select the corresponding of the k outputs of all k with a selection block.
 *
 * @author Thomas Schneider
 */
public class UCsel extends UC {
	private static final long serialVersionUID = 1L;
	
	UCsel UC=this;
	
	private Ublock ub;
	private S_u_v sin;
	private S_u_v sout;	
	
	public int countGates() {	
		int cin=u;
		int csi=sin.countGates();
		int cu=ub.countGates();
		int cso=sout.countGates();
		int csum=cin+csi+cu+cso;
		
		logger.debug("sizeof(inputs)=\t"+cin+"\t("+(cin*100)/csum+" %)");
		logger.debug("sizeof(SI)=\t"+csi+"\t("+(csi*100)/csum+" %)");
		logger.debug("sizeof(U)=\t"+cu+"\t("+(cu*100)/csum+" %)");
		logger.debug("sizeof(SO)=\t"+cso+"\t("+(cso*100)/csum+" %)");
		logger.debug("sum=\t\t"+csum);
		
		return csum;		
	}

	public int estimateGates(int ins, int outs, int k) {
		return k * ( k + 2*ins + outs - 2 ) - outs;
	}

	public Gate[] extractCircuit(Gate[] outputs) {		
		Gate si_out[];
		Gate u_out[];
		Gate inputs[];
		if(outputs==null) outputs=new Gate[v];

		logger.info("Extracting circuit representation of UC.");

		// create input k
		inputs=new Gate[u];
		for(int i=0; i<u; i++)
			inputs[i]=addGate(1, Gate.INP_GATE, false);

		//logger.debug("SI---");
		si_out=sin.extractCircuit(inputs,null);
		//logger.debug("U---");
		u_out=ub.extractCircuit(si_out,null);
		//logger.debug("SO---");
		outputs=sout.extractCircuit(u_out, outputs);
		
		// mark outputs
		for(int i=0; i<outputs.length; i++)
			outputs[i].setOutput(true);
		
		// get output indexes
		output_indexes=new int[v];
		for(int i=0; i<v; i++) output_indexes[i]=outputs[i].gate_index;
		
		return outputs;
	}

	public void generate(int u, int v, int k) {
		logger.info("Generating UCsel for u="+u+", v="+v+", k="+k+".");
		
		if (v==0) {
			logger.error("Circuit must have at least 1 output.");
			System.exit(1);
		}
		
		this.u=u;
		this.v=v;
		this.k=k;

		sin=new S_u_v(u,2*k,UC);
		ub=new Ublock();
		sout=new S_u_v(k, v, UC);

		is_generated=true;
		
	}	
	
	public void program(int[] ins, int[] outs, ParseGate[] parsegates) {		
		int sinprog[];
		int soutprog[];
		
		logger.info("Programming UC.");
		
		// program U
		sinprog=new int[2*k];
		ub.program(parsegates,sinprog);

		// program Sin
		for(int i=0; i<sinprog.length; i++) {
			sinprog[i]=translateIn[sinprog[i]];
		}
		sin.program(sinprog);
		
		// program Sout
		soutprog=new int[v];
		for(int i=0; i<soutprog.length; i++) {
			soutprog[i]=translateOut[i]-u;
		}
		sout.program(soutprog);	
	}
	
	
	// Universal Block ----------------------------------------
	class Ublock {
		S_u_1 s[];
		G g[];

		void program(ParseGate gp[], int sinprog[]) {			
			int aktin;
			for(int i=0; i<k; i++) {
				ParseGate aktg=gp[i];

				g[i].program(aktg.truthTable);

				// left
				aktin=aktg.inputs[0];
				if(aktin<u) {
					sinprog[2*i] =aktin;
					s[2*i].program(0);
				}
				else {
					sinprog[2*i] =0;					
					s[2*i].program(aktin-u+1);
				}
				
				// right
				aktin=aktg.inputs[1];
				if(aktin<u) {
					sinprog[2*i+1] =aktin;
					s[2*i+1].program(0);
				}
				else {
					sinprog[2*i+1] =0;					
					s[2*i+1].program(aktin-u+1);
				}
			}
		}
		
		Gate[] extractCircuit(Gate[] inputs, Gate[] outputs) {
			if(outputs==null) outputs=new Gate[k];
			
			Gate out1[]=new Gate[1];
			Gate out2[]=new Gate[1];
			Gate in[];
			for(int i=0; i<k; i++) {
				in=new Gate[i+1];
				for(int j=0; j<i; j++) in[j+1]=outputs[j];
				
				// left
				in[0]=inputs[2*i];
				out1=s[2*i].extractCircuit(in, out1);
				
				// right
				in[0]=inputs[2*i+1];
				out2=s[2*i+1].extractCircuit(in, out2);
				
				outputs[i]=g[i].extractCircuit(out1[0], out2[0]);
			}
			return outputs;
		}

		int countGates() {
			int Scount=0;
			for(int i=0; i<k-1; i++) Scount+=s[2*i].countGates();
			return g.length+Scount;
		}
		
		Ublock() {
			g=new G[k];
			s=new S_u_1[2*k];

			for(int i=0; i<g.length; i++) g[i]=new G();
			for(int i=0; i<k; i++) {
				s[2*i]  =new S_u_1(i+1,UC);
				s[2*i+1]=new S_u_1(i+1,UC);				
			}
		}
	}
	
	public UCsel() {
		is_implemented=true;
	}
}
